@@ -22,27 +22,29 @@
 """Generic utility helpers."""
 from __future__ import annotations
 
+import ast
 import contextlib
 import inspect
 import logging
 import os
 import re
-from collections.abc import Generator, ItemsView, Iterator, Mapping, Sequence
+from collections.abc import ItemsView, Iterator, Mapping, Sequence
 from dataclasses import _MISSING_TYPE, dataclass, field
-from functools import cache
+from functools import cache, lru_cache
 from pathlib import Path
-from typing import Any, Callable
+from typing import TYPE_CHECKING, Any
 
+import ruamel.yaml.parser
 import yaml
 from ansible.errors import AnsibleError, AnsibleParserError
 from ansible.module_utils.parsing.convert_bool import boolean
 from ansible.parsing.dataloader import DataLoader
 from ansible.parsing.mod_args import ModuleArgsParser
-from ansible.parsing.splitter import split_args
+from ansible.parsing.plugin_docs import read_docstring
 from ansible.parsing.yaml.constructor import AnsibleConstructor, AnsibleMapping
 from ansible.parsing.yaml.loader import AnsibleLoader
 from ansible.parsing.yaml.objects import AnsibleBaseYAMLObject, AnsibleSequence
-from ansible.plugins.loader import add_all_plugin_dirs
+from ansible.plugins.loader import PluginLoadContext, add_all_plugin_dirs, module_loader
 from ansible.template import Templar
 from ansible.utils.collection_loader import AnsibleCollectionConfig
 from yaml.composer import Composer
@@ -50,12 +52,12 @@ from yaml.representer import RepresenterError
 
 from ansiblelint._internal.rules import (
     AnsibleParserErrorRule,
-    LoadingFailureRule,
     RuntimeErrorRule,
 )
 from ansiblelint.app import get_app
 from ansiblelint.config import Options, options
 from ansiblelint.constants import (
+    ANNOTATION_KEYS,
     FILENAME_KEY,
     INCLUSION_ACTION_NAMES,
     LINE_NUMBER_KEY,
@@ -68,8 +70,10 @@ from ansiblelint.constants import (
 from ansiblelint.errors import MatchError
 from ansiblelint.file_utils import Lintable, discover_lintables
 from ansiblelint.skip_utils import is_nested_task
-from ansiblelint.text import removeprefix
+from ansiblelint.text import has_jinja, removeprefix
 
+if TYPE_CHECKING:
+    from ansiblelint.rules import RulesCollection
 # ansible-lint doesn't need/want to know about encrypted secrets, so we pass a
 # string as the password to enable such yaml files to be opened and parsed
 # successfully.
@@ -187,10 +191,7 @@ def ansible_template(
                     _logger.warning(err)
                     raise
 
-                # pylint: disable=protected-access
-                templar.environment.filters._delegatee[  # noqa: SLF001
-                    missing_filter
-                ] = mock_filter
+                templar.environment.filters._delegatee[missing_filter] = mock_filter  # fmt: skip # noqa: SLF001
                 # Record the mocked filter so we can warn the user
                 if missing_filter not in options.mock_filters:
                     _logger.debug("Mocking missing filter %s", missing_filter)
@@ -233,7 +234,8 @@ def tokenize(line: str) -> tuple[str, list[str], dict[str, str]]:
     return (command, args, kwargs)
 
 
-def _playbook_items(pb_data: AnsibleBaseYAMLObject) -> ItemsView:  # type: ignore[type-arg]
+def playbook_items(pb_data: AnsibleBaseYAMLObject) -> ItemsView:  # type: ignore[type-arg]
+    """Return a list of items from within the playbook."""
     if isinstance(pb_data, dict):
         return pb_data.items()
     if not pb_data:
@@ -245,62 +247,13 @@ def _playbook_items(pb_data: AnsibleBaseYAMLObject) -> ItemsView:  # type: ignor
     return [item for play in pb_data if play for item in play.items()]  # type: ignore[return-value]
 
 
-def _set_collections_basedir(basedir: Path) -> None:
-    # Sets the playbook directory as playbook_paths for the collection loader
+def set_collections_basedir(basedir: Path) -> None:
+    """Set the playbook directory as playbook_paths for the collection loader."""
     # Ansible expects only absolute paths inside `playbook_paths` and will
     # produce weird errors if we use a relative one.
     AnsibleCollectionConfig.playbook_paths = str(basedir.resolve())
 
 
-def find_children(lintable: Lintable) -> list[Lintable]:  # noqa: C901
-    """Traverse children of a single file or folder."""
-    if not lintable.path.exists():
-        return []
-    playbook_dir = str(lintable.path.parent)
-    _set_collections_basedir(lintable.path.parent)
-    add_all_plugin_dirs(playbook_dir or ".")
-    if lintable.kind == "role":
-        playbook_ds = AnsibleMapping({"roles": [{"role": str(lintable.path)}]})
-    elif lintable.kind not in ("playbook", "tasks"):
-        return []
-    else:
-        try:
-            playbook_ds = parse_yaml_from_file(str(lintable.path))
-        except AnsibleError as exc:
-            raise SystemExit(exc) from exc
-    results = []
-    # playbook_ds can be an AnsibleUnicode string, which we consider invalid
-    if isinstance(playbook_ds, str):
-        raise MatchError(lintable=lintable, rule=LoadingFailureRule())
-    for item in _playbook_items(playbook_ds):
-        # if lintable.kind not in ["playbook"]:
-        for child in play_children(
-            lintable.path.parent,
-            item,
-            lintable.kind,
-            playbook_dir,
-        ):
-            # We avoid processing parametrized children
-            path_str = str(child.path)
-            if "$" in path_str or "{{" in path_str:
-                continue
-
-            # Repair incorrect paths obtained when old syntax was used, like:
-            # - include: simpletask.yml tags=nginx
-            valid_tokens = []
-            for token in split_args(path_str):
-                if "=" in token:
-                    break
-                valid_tokens.append(token)
-            path = " ".join(valid_tokens)
-            if path != path_str:
-                child.path = Path(path)
-                child.name = child.path.name
-
-            results.append(child)
-    return results
-
-
 def template(
     basedir: Path,
     value: Any,
@@ -318,7 +271,7 @@ def template(
             variables,
             **dict(kwargs, fail_on_undefined=fail_on_undefined),
         )
-        # Hack to skip the following exception when using to_json filter on a variable.
+        # Hack to skip the following exception when using to_json filter on a variable. # noqa: FIX004
         # I guess the filter doesn't like empty vars...
     except (AnsibleError, ValueError, RepresenterError):
         # templating failed, so just keep value as is.
@@ -327,145 +280,169 @@ def template(
     return value
 
 
-def play_children(
-    basedir: Path,
-    item: tuple[str, Any],
-    parent_type: FileType,
-    playbook_dir: str,  # noqa: ARG001
-) -> list[Lintable]:
-    """Flatten the traversed play tasks."""
-    # pylint: disable=unused-argument
-    delegate_map: dict[str, Callable[[str, Any, Any, FileType], list[Lintable]]] = {
-        "tasks": _taskshandlers_children,
-        "pre_tasks": _taskshandlers_children,
-        "post_tasks": _taskshandlers_children,
-        "block": _taskshandlers_children,
-        "include": _include_children,
-        "ansible.builtin.include": _include_children,
-        "import_playbook": _include_children,
-        "ansible.builtin.import_playbook": _include_children,
-        "roles": _roles_children,
-        "dependencies": _roles_children,
-        "handlers": _taskshandlers_children,
-        "include_tasks": _include_children,
-        "ansible.builtin.include_tasks": _include_children,
-        "import_tasks": _include_children,
-        "ansible.builtin.import_tasks": _include_children,
-    }
-    (k, v) = item
-    add_all_plugin_dirs(str(basedir.resolve()))
+@dataclass
+class HandleChildren:
+    """Parse task, roles and children."""
 
-    if k in delegate_map and v:
-        v = template(
-            basedir,
-            v,
-            {"playbook_dir": PLAYBOOK_DIR or str(basedir.resolve())},
-            fail_on_undefined=False,
-        )
-        return delegate_map[k](str(basedir), k, v, parent_type)
-    return []
+    rules: RulesCollection = field(init=True, repr=False)
 
+    def include_children(
+        self,
+        basedir: str,
+        k: str,
+        v: Any,
+        parent_type: FileType,
+    ) -> list[Lintable]:
+        """Include children."""
+        # handle special case include_tasks: name=filename.yml
+        if k in INCLUSION_ACTION_NAMES and isinstance(v, dict) and "file" in v:
+            v = v["file"]
 
-def _include_children(
-    basedir: str,
-    k: str,
-    v: Any,
-    parent_type: FileType,
-) -> list[Lintable]:
-    # handle special case include_tasks: name=filename.yml
-    if k in INCLUSION_ACTION_NAMES and isinstance(v, dict) and "file" in v:
-        v = v["file"]
+        # we cannot really parse any jinja2 in includes, so we ignore them
+        if not v or "{{" in v:
+            return []
 
-    # we cannot really parse any jinja2 in includes, so we ignore them
-    if not v or "{{" in v:
-        return []
+        if "import_playbook" in k and COLLECTION_PLAY_RE.match(v):
+            # Any import_playbooks from collections should be ignored as ansible
+            # own syntax check will handle them.
+            return []
 
-    if "import_playbook" in k and COLLECTION_PLAY_RE.match(v):
-        # Any import_playbooks from collections should be ignored as ansible
-        # own syntax check will handle them.
-        return []
+        # handle include: filename.yml tags=blah
+        # pylint: disable=unused-variable
+        (command, args, kwargs) = tokenize(f"{k}: {v}")
 
-    # handle include: filename.yml tags=blah
-    # pylint: disable=unused-variable
-    (command, args, kwargs) = tokenize(f"{k}: {v}")
-
-    result = path_dwim(basedir, args[0])
-    while basedir not in ["", "/"]:
-        if os.path.exists(result):
-            break
-        basedir = os.path.dirname(basedir)
         result = path_dwim(basedir, args[0])
+        while basedir not in ["", "/"]:
+            if os.path.exists(result):
+                break
+            basedir = os.path.dirname(basedir)
+            result = path_dwim(basedir, args[0])
 
-    return [Lintable(result, kind=parent_type)]
+        return [Lintable(result, kind=parent_type)]
 
-
-def _taskshandlers_children(
-    basedir: str,
-    k: str,
-    v: None | Any,
-    parent_type: FileType,
-) -> list[Lintable]:
-    results: list[Lintable] = []
-    if v is None:
-        raise MatchError(
-            message="A malformed block was encountered while loading a block.",
-            rule=RuntimeErrorRule(),
-        )
-    for task_handler in v:
-        # ignore empty tasks, `-`
-        if not task_handler:
-            continue
-
-        with contextlib.suppress(LookupError):
-            children = _get_task_handler_children_for_tasks_or_playbooks(
-                task_handler,
-                basedir,
-                k,
-                parent_type,
+    def taskshandlers_children(
+        self,
+        basedir: str,
+        k: str,
+        v: None | Any,
+        parent_type: FileType,
+    ) -> list[Lintable]:
+        """TasksHandlers Children."""
+        results: list[Lintable] = []
+        if v is None:
+            raise MatchError(
+                message="A malformed block was encountered while loading a block.",
+                rule=RuntimeErrorRule(),
             )
-            results.append(children)
-            continue
+        for task_handler in v:
+            # ignore empty tasks, `-`
+            if not task_handler:
+                continue
 
-        if any(x in task_handler for x in ROLE_IMPORT_ACTION_NAMES):
-            task_handler = normalize_task_v2(task_handler)
-            _validate_task_handler_action_for_role(task_handler["action"])
-            results.extend(
-                _roles_children(
+            with contextlib.suppress(LookupError):
+                children = _get_task_handler_children_for_tasks_or_playbooks(
+                    task_handler,
                     basedir,
                     k,
-                    [task_handler["action"].get("name")],
                     parent_type,
-                    main=task_handler["action"].get("tasks_from", "main"),
-                ),
-            )
-            continue
+                )
+                results.append(children)
+                continue
 
-        if "block" not in task_handler:
-            continue
+            if any(x in task_handler for x in ROLE_IMPORT_ACTION_NAMES):
+                task_handler = normalize_task_v2(task_handler)
+                self._validate_task_handler_action_for_role(task_handler["action"])
+                name = task_handler["action"].get("name")
+                if has_jinja(name):
+                    # we cannot deal with dynamic imports
+                    continue
+                results.extend(
+                    self.roles_children(basedir, k, [name], parent_type),
+                )
+                continue
+
+            if "block" not in task_handler:
+                continue
 
-        results.extend(
-            _taskshandlers_children(basedir, k, task_handler["block"], parent_type),
-        )
-        if "rescue" in task_handler:
             results.extend(
-                _taskshandlers_children(
+                self.taskshandlers_children(
                     basedir,
                     k,
-                    task_handler["rescue"],
+                    task_handler["block"],
                     parent_type,
                 ),
             )
-        if "always" in task_handler:
-            results.extend(
-                _taskshandlers_children(
-                    basedir,
-                    k,
-                    task_handler["always"],
-                    parent_type,
+            if "rescue" in task_handler:
+                results.extend(
+                    self.taskshandlers_children(
+                        basedir,
+                        k,
+                        task_handler["rescue"],
+                        parent_type,
+                    ),
+                )
+            if "always" in task_handler:
+                results.extend(
+                    self.taskshandlers_children(
+                        basedir,
+                        k,
+                        task_handler["always"],
+                        parent_type,
+                    ),
+                )
+
+        return results
+
+    def _validate_task_handler_action_for_role(self, th_action: dict[str, Any]) -> None:
+        """Verify that the task handler action is valid for role include."""
+        module = th_action["__ansible_module__"]
+
+        if "name" not in th_action:
+            raise MatchError(
+                message=f"Failed to find required 'name' key in {module!s}",
+                rule=self.rules.rules[0],
+                filename=(
+                    self.rules.options.lintables[0]
+                    if self.rules.options.lintables
+                    else "."
                 ),
             )
 
-    return results
+        if not isinstance(th_action["name"], str):
+            raise MatchError(
+                message=f"Value assigned to 'name' key on '{module!s}' is not a string.",
+                rule=self.rules.rules[1],
+            )
+
+    def roles_children(
+        self,
+        basedir: str,
+        k: str,
+        v: Sequence[Any],
+        parent_type: FileType,
+    ) -> list[Lintable]:
+        """Roles children."""
+        # pylint: disable=unused-argument # parent_type)
+        results: list[Lintable] = []
+        if not v:
+            # typing does not prevent junk from being passed in
+            return results
+        for role in v:
+            if isinstance(role, dict):
+                if "role" in role or "name" in role:
+                    if "tags" not in role or "skip_ansible_lint" not in role["tags"]:
+                        results.extend(
+                            _look_for_role_files(
+                                basedir,
+                                role.get("role", role.get("name")),
+                            ),
+                        )
+                elif k != "dependencies":
+                    msg = f'role dict {role} does not contain a "role" or "name" key'
+                    raise SystemExit(msg)
+            else:
+                results.extend(_look_for_role_files(basedir, role))
+        return results
 
 
 def _get_task_handler_children_for_tasks_or_playbooks(
@@ -499,50 +476,6 @@ def _get_task_handler_children_for_tasks_or_playbooks(
     raise LookupError(msg)
 
 
-def _validate_task_handler_action_for_role(th_action: dict[str, Any]) -> None:
-    """Verify that the task handler action is valid for role include."""
-    module = th_action["__ansible_module__"]
-
-    if "name" not in th_action:
-        raise MatchError(message=f"Failed to find required 'name' key in {module!s}")
-
-    if not isinstance(th_action["name"], str):
-        raise MatchError(
-            message=f"Value assigned to 'name' key on '{module!s}' is not a string.",
-        )
-
-
-def _roles_children(
-    basedir: str,
-    k: str,
-    v: Sequence[Any],
-    parent_type: FileType,  # noqa: ARG001
-    main: str = "main",
-) -> list[Lintable]:
-    # pylint: disable=unused-argument # parent_type)
-    results: list[Lintable] = []
-    if not v:
-        # typing does not prevent junk from being passed in
-        return results
-    for role in v:
-        if isinstance(role, dict):
-            if "role" in role or "name" in role:
-                if "tags" not in role or "skip_ansible_lint" not in role["tags"]:
-                    results.extend(
-                        _look_for_role_files(
-                            basedir,
-                            role.get("role", role.get("name")),
-                            main=main,
-                        ),
-                    )
-            elif k != "dependencies":
-                msg = f'role dict {role} does not contain a "role" or "name" key'
-                raise SystemExit(msg)
-        else:
-            results.extend(_look_for_role_files(basedir, role, main=main))
-    return results
-
-
 def _rolepath(basedir: str, role: str) -> str | None:
     role_path = None
 
@@ -557,7 +490,7 @@ def _rolepath(basedir: str, role: str) -> str | None:
         path_dwim(basedir, os.path.join("..", role)),
     ]
 
-    for loc in get_app().runtime.config.default_roles_path:
+    for loc in get_app(offline=True).runtime.config.default_roles_path:
         loc = os.path.expanduser(loc)
         possible_paths.append(path_dwim(loc, role))
 
@@ -574,12 +507,7 @@ def _rolepath(basedir: str, role: str) -> str | None:
     return role_path
 
 
-def _look_for_role_files(
-    basedir: str,
-    role: str,
-    main: str | None = "main",  # noqa: ARG001
-) -> list[Lintable]:
-    # pylint: disable=unused-argument # main
+def _look_for_role_files(basedir: str, role: str) -> list[Lintable]:
     role_path = _rolepath(basedir, role)
     if not role_path:  # pragma: no branch
         return []
@@ -650,7 +578,6 @@ def normalize_task_v2(task: dict[str, Any]) -> dict[str, Any]:
             skip_action_validation=options.skip_action_validation,
         )
     except AnsibleParserError as exc:
-        # pylint: disable=raise-missing-from
         raise MatchError(
             rule=AnsibleParserErrorRule(),
             message=exc.message,
@@ -722,7 +649,7 @@ def task_to_str(task: dict[str, Any]) -> str:
     _raw_params = action.get("_raw_params", [])
     if isinstance(_raw_params, list):
         for item in _raw_params:
-            args.append(str(item))
+            args.extend(str(item))
     else:
         args.append(_raw_params)
 
@@ -786,7 +713,35 @@ class Task(dict[str, Any]):
     @property
     def name(self) -> str | None:
         """Return the name of the task."""
-        return self.raw_task.get("name", None)
+        name = self.raw_task.get("name", None)
+        if name is not None and not isinstance(name, str):
+            msg = "Task name can only be a string."
+            raise RuntimeError(msg)
+        return name
+
+    @property
+    def action(self) -> str:
+        """Return the resolved action name."""
+        action_name = self.normalized_task["action"]["__ansible_module_original__"]
+        if not isinstance(action_name, str):
+            msg = "Task actions can only be strings."
+            raise RuntimeError(msg)
+        return action_name
+
+    @property
+    def args(self) -> Any:
+        """Return the arguments passed to the task action.
+
+        While we usually expect to return a dictionary, it can also
+        return a templated string when jinja is used.
+        """
+        if "args" in self.raw_task:
+            return self.raw_task["args"]
+        result = {}
+        for k, v in self.normalized_task["action"].items():
+            if k not in ANNOTATION_KEYS:
+                result[k] = v
+        return result
 
     @property
     def normalized_task(self) -> dict[str, Any]:
@@ -813,6 +768,15 @@ class Task(dict[str, Any]):
         skip_tags: list[str] = self.raw_task.get(SKIPPED_RULES_KEY, [])
         return skip_tags
 
+    def is_handler(self) -> bool:
+        """Return true for tasks that are handlers."""
+        is_handler_file = False
+        if isinstance(self._normalized_task, dict):
+            file_name = str(self._normalized_task["action"].get(FILENAME_KEY, None))
+            if file_name:
+                is_handler_file = "handlers" in str(file_name)
+        return is_handler_file if is_handler_file else ".handlers[" in self.position
+
     def __repr__(self) -> str:
         """Return a string representation of the task."""
         return f"Task('{self.name}' [{self.position}])"
@@ -825,12 +789,12 @@ class Task(dict[str, Any]):
         """Allow access as task[...]."""
         return self.normalized_task[index]
 
-    def __iter__(self) -> Generator[str, None, None]:
+    def __iter__(self) -> Iterator[str]:
         """Provide support for 'key in task'."""
         yield from (f for f in self.normalized_task)
 
 
-def task_in_list(  # noqa: C901
+def task_in_list(
     data: AnsibleBaseYAMLObject,
     file: Lintable,
     kind: str,
@@ -906,7 +870,7 @@ def get_action_tasks(data: AnsibleBaseYAMLObject, file: Lintable) -> list[Any]:
 
 
 @cache
-def parse_yaml_linenumbers(  # noqa: max-complexity: 12
+def parse_yaml_linenumbers(
     lintable: Lintable,
 ) -> AnsibleBaseYAMLObject:
     """Parse yaml as ansible.utils.parse_yaml but with linenumbers.
@@ -934,9 +898,7 @@ def parse_yaml_linenumbers(  # noqa: max-complexity: 12
         if hasattr(node, "__line__"):
             mapping[LINE_NUMBER_KEY] = node.__line__
         else:
-            mapping[
-                LINE_NUMBER_KEY
-            ] = mapping._line_number  # pylint: disable=protected-access  # noqa: SLF001
+            mapping[LINE_NUMBER_KEY] = mapping._line_number  # noqa: SLF001
         mapping[FILENAME_KEY] = lintable.path
         return mapping
 
@@ -959,8 +921,9 @@ def parse_yaml_linenumbers(  # noqa: max-complexity: 12
         yaml.parser.ParserError,
         yaml.scanner.ScannerError,
         yaml.constructor.ConstructorError,
+        ruamel.yaml.parser.ParserError,
     ) as exc:
-        msg = "Failed to load YAML file"
+        msg = f"Failed to load YAML file: {lintable.path}"
         raise RuntimeError(msg) from exc
 
     if len(result) == 0:
@@ -1039,7 +1002,6 @@ def is_playbook(filename: str) -> bool:
     return False
 
 
-# pylint: disable=too-many-statements
 def get_lintables(
     opts: Options = options,
     args: list[str] | None = None,
@@ -1082,3 +1044,43 @@ def _extend_with_roles(lintables: list[Lintable]) -> None:
 def convert_to_boolean(value: Any) -> bool:
     """Use Ansible to convert something to a boolean."""
     return bool(boolean(value))
+
+
+def parse_examples_from_plugin(lintable: Lintable) -> tuple[int, str]:
+    """Parse yaml inside plugin EXAMPLES string.
+
+    Store a line number offset to realign returned line numbers later
+    """
+    offset = 1
+    parsed = ast.parse(lintable.content)
+    for child in parsed.body:
+        if isinstance(child, ast.Assign):
+            label = child.targets[0]
+            if isinstance(label, ast.Name) and label.id == "EXAMPLES":
+                offset = child.lineno - 1
+                break
+
+    docs = read_docstring(str(lintable.path))
+    examples = docs["plainexamples"]
+
+    # Ignore the leading newline and lack of document start
+    # as including those in EXAMPLES would be weird.
+    return offset, (f"---{examples}" if examples else "")
+
+
+@lru_cache
+def load_plugin(name: str) -> PluginLoadContext:
+    """Return loaded ansible plugin/module."""
+    loaded_module = module_loader.find_plugin_with_context(
+        name,
+        ignore_deprecated=True,
+        check_aliases=True,
+    )
+    if not loaded_module.resolved and name.startswith("ansible.builtin."):
+        # fallback to core behavior of using legacy
+        loaded_module = module_loader.find_plugin_with_context(
+            name.replace("ansible.builtin.", "ansible.legacy."),
+            ignore_deprecated=True,
+            check_aliases=True,
+        )
+    return loaded_module
